@@ -1,9 +1,11 @@
-
 from lisa.utils import LoadingBar
 from multiprocessing import Pool
 import numpy as np
 from scipy import stats, sparse
 import os
+#from pyranges.pyranges import PyRange
+from lisa.models import UnweightedChromatinModel
+from collections import defaultdict
 
 def get_delta_RP(profile, binding_data, rp_map):
     '''
@@ -33,18 +35,26 @@ def mannu_test_function(x):
 
 class LISA_RP_Assay:
 
-    invitro_metadata = ['factor','cell_line','cell_type','tissue']
-    insilico_metadata = ['factor','edition','source','sourcefile','status','numseqs','pmid']
-
-    def __init__(self, *, technology, config, cores, log, metadata_headers, metadata_path, oneshot):
+    def __init__(self, *, technology, config, cores, log, metadata):
         self.config = config
         self.technology = technology
         self.cores = cores
         self.log = log
         self.loaded = False
-        self.oneshot = False
-        self.metadata_headers = metadata_headers
-        self.metadata_path = metadata_path
+        self.metadata = metadata
+
+    @staticmethod
+    def make_subset_rp_map(*,rp_map, factor_binding, gene_mask, accessibility):
+        
+        bin_mask = np.squeeze(np.array(rp_map[gene_mask, : ].tocsc().sum(axis = 0) > 0))
+        #subset rp_map and factor hits on bins with RP > 0
+        subset_factor_binding = factor_binding[bin_mask, :]
+        
+        subset_rp_map = rp_map[gene_mask, :][:, bin_mask]
+
+        subset_accessibility = accessibility[bin_mask, :]
+        
+        return subset_accessibility, subset_factor_binding, subset_rp_map
 
     def load_rp_matrix(self, data_object, gene_mask = None):
 
@@ -52,11 +62,7 @@ class LISA_RP_Assay:
         
         dataset_ids = data_object[self.config.get('accessibility_assay', 'reg_potential_dataset_ids').format(technology = self.technology)][...].astype(str)
 
-        if self.oneshot and not gene_mask is None:
-            rp_matrix = data_object[self.config.get('accessibility_assay', 'reg_potential_matrix').format(technology = self.technology)][gene_mask, :]
-        else:
-            self.oneshot = False
-            rp_matrix = data_object[self.config.get('accessibility_assay', 'reg_potential_matrix').format(technology = self.technology)][...]
+        rp_matrix = data_object[self.config.get('accessibility_assay', 'reg_potential_matrix').format(technology = self.technology)][...]
         
         self.loaded = True
         return rp_matrix, dataset_ids
@@ -91,27 +97,8 @@ class LISA_RP_Assay:
     def predict(self, gene_mask, label_vector, data_object = None, debug = False):
         raise NotImplementedError()
 
-    def load_metadata(self):
-        #reformat metadata tsv into conventiant id-indexed dictionary
-        with open(self.metadata_path, 'r', encoding = 'latin') as metdata_file:
-            metadata = [[field.strip() for field in line.split('\t')] for line in metdata_file.readlines()]
-                                
-        meta_headers, metadata = metadata[0], metadata[1:]
-
-        return {metaline[0] : dict(zip(meta_headers[1:], metaline[1:])) for metaline in metadata}
-
     def load(self, data_object, gene_mask = None):
         self.rp_matrix, self.dataset_ids = self.load_rp_matrix(data_object, gene_mask = gene_mask)
-        self.metadata = self.load_metadata()
-
-    def get_metadata(self, sample_ids):
-        return dict(
-            sample_id = sample_ids,
-            **{
-                header : [self.metadata[_id][header] for _id in sample_ids]
-                for header in self.metadata_headers 
-            }
-        )
 
 class PeakRP_Assay(LISA_RP_Assay):
 
@@ -125,16 +112,13 @@ class PeakRP_Assay(LISA_RP_Assay):
         self.log.append('Calculating {} peak-RP p-values ...'.format(self.technology))
         
         #calculate p-values by directly applying mannu-test on RP matrix. Subset the RP matrix for genes-of-interest if required
-        p_vals = self.get_delta_RP_p_value(self.rp_matrix[gene_mask, :] if not self.oneshot else self.rp_matrix, label_vector)
+        p_vals = self.get_delta_RP_p_value(self.rp_matrix[gene_mask, :], label_vector)
 
         return p_vals
 
     def get_info(self):
         return dict()
 
-    def get_metadata(self):
-        return super().get_metadata(self.dataset_ids)
-
 #Generator that repeatedly yields the same factor_binding and rp_map matrices with a new accessibility profile 
 #to the multiprocessing pool creator. This reduces data redundancy.
 class KnockoutGenerator:
@@ -150,20 +134,22 @@ class KnockoutGenerator:
     
 class Accesibility_Assay(LISA_RP_Assay):
 
-    def __init__(self, *, technology, config, cores, log, metadata_path, oneshot, rp_map, factor_binding, selection_model, chromatin_model,
-        metadata_headers = LISA_RP_Assay.invitro_metadata):
-        super().__init__(technology = technology, config=config,cores=cores, log=log, metadata_headers=metadata_headers,
-            metadata_path=metadata_path, oneshot=oneshot)
+    def __init__(self, *, technology, config, cores, log, rp_map, factor_binding, selection_model, chromatin_model,
+        metadata, factor_gene_mask):
+        super().__init__(technology = technology, config=config,cores=cores, log=log,
+            metadata = metadata)
         self.rp_map = rp_map
         self.factor_binding = factor_binding
         self.selection_model = selection_model
         self.chromatin_model = chromatin_model
+        self.factor_gene_mask = factor_gene_mask
 
     def get_info(self):
         return dict(
             selection_model = self.selection_model.get_info(),
             chromatin_model = self.chromatin_model.get_info(),
-            selected_datasets = self.get_metadata(list(self.selected_dataset_ids)),
+            selected_datasets = self.metadata.select(list(self.selected_dataset_ids)),
+            factor_acc_z_scores = self.factor_acc_z_scores
         )
 
     def load_accessibility_profiles(self, data_object, selected_dataset_ids):
@@ -187,7 +173,6 @@ class Accesibility_Assay(LISA_RP_Assay):
 
     def calculate_ISDs(self, accessibility_profiles, factor_binding, rp_map): #enforced kwargs
         """
-        log: log object for printing
         accessibility_profiles: (bins, num_datasets): list of chromatin-accessiblity datasets on which to analyze the effects of ISD
         factor_binding: (num_bins, TFs), a sparse binary map showing bins that contain a chip-seq or motif peak to knock out
         rp_map: (num_bins, genes), precalculated matrix mapping the reads in a bin to a regulatory score for each gene (this is a huge matrix)
@@ -195,15 +180,6 @@ class Accesibility_Assay(LISA_RP_Assay):
         returns:
         delta_regulatory_score (genes x TFs)
         """
-        '''
-        # Code for non-multiprocessing implementation of this ISD algorithm. Much slower.
-        knockouts = []
-        loadingbar = LoadingBar('Performing in-silico knockouts', accessibility_profiles.shape[1], 20, cold_start = True)
-        log.append(loadingbar, update_line = True)
-        for profile in accessibility_profiles.T:
-            knockouts.append(get_delta_RP(profile, factor_binding, rp_map))
-            log.append(loadingbar, update_line = True)'''
-
         self.log.append('Performing in-silico knockouts ...')
 
         with Pool(self.cores) as p:
@@ -221,7 +197,27 @@ class Accesibility_Assay(LISA_RP_Assay):
 
         return delta_regulation_score, datacube
 
-    
+    def introsect_accessibility(self, rp_matrix, gene_mask, dataset_mask, dataset_coefs):
+
+        #print(rp_matrix.shape, gene_mask.shape, gene_mask.dtype)
+
+        factor_accessiblities = np.log2(rp_matrix[gene_mask, :] + 1)
+
+        background_datasets = factor_accessiblities[:, ~dataset_mask]
+
+        factor_means, factor_stds = background_datasets.mean(axis = 1, keepdims = True), background_datasets.std(axis = 1, keepdims = True)
+        factor_acc_z_scores = (factor_accessiblities[:, dataset_mask] - factor_means)/factor_stds
+
+        dataset_coefs = dataset_coefs.reshape(-1)
+        dataset_coefs_mask = dataset_coefs > 0
+        if dataset_coefs_mask.sum() == 0:
+            dataset_coefs_mask = np.ones_like(dataset_coefs).astype(np.bool)
+
+        coef_weights = dataset_coefs[dataset_coefs_mask]/dataset_coefs[dataset_coefs_mask].sum()
+        factor_acc_z_scores = np.mean(np.multiply(factor_acc_z_scores[:, dataset_coefs_mask], coef_weights.reshape(1,-1)), axis = 1)
+
+        return factor_acc_z_scores
+
     def predict(self, gene_mask, label_vector, data_object, debug = False):
 
         with self.log.section('Modeling {} purturbations:'.format(self.technology)):
@@ -231,14 +227,7 @@ class Accesibility_Assay(LISA_RP_Assay):
             except AttributeError:
                 self.load(data_object, gene_mask= gene_mask)
 
-            #find bins of gene-subsetted rp-map with RP > 0
-            bin_mask = np.squeeze(np.array(self.rp_map[gene_mask, : ].tocsc().sum(axis = 0) > 0))
-            #subset rp_map and factor hits on bins with RP > 0
-            subset_factor_binding = self.factor_binding[bin_mask, :]
-            
-            subset_rp_map = self.rp_map[gene_mask, :][:, bin_mask]
-
-            subset_rp_matrix = self.rp_matrix[gene_mask, :] if not self.oneshot else self.rp_matrix
+            subset_rp_matrix = self.rp_matrix[gene_mask, :]
 
             #DNase model building and purturbation
             self.log.append('Selecting discriminative datasets and training chromatin model ...')
@@ -246,6 +235,7 @@ class Accesibility_Assay(LISA_RP_Assay):
             #select the most discriminative datasets
             dataset_mask = self.selection_model.fit(subset_rp_matrix, label_vector)
             #subset the best datasets
+
             subset_rp_matrix, self.selected_dataset_ids = subset_rp_matrix[:, dataset_mask], self.dataset_ids[dataset_mask]
             #fit the chromatin model to these datasets
             self.chromatin_model.fit(subset_rp_matrix, label_vector, self.cores)
@@ -254,14 +244,20 @@ class Accesibility_Assay(LISA_RP_Assay):
 
                 accesibility_profiles = self.load_accessibility_profiles(data_object, self.selected_dataset_ids)
 
-                accesibility_profiles = accesibility_profiles[bin_mask, :]
+                subset_accessibility, subset_factor_binding, subset_rp_map = self.make_subset_rp_map(rp_map = self.rp_map, 
+                    factor_binding = self.factor_binding, gene_mask = gene_mask, accessibility = accesibility_profiles)
                 
-                delta_reg_scores, datacube = self.calculate_ISDs(accesibility_profiles, subset_factor_binding, subset_rp_map)
+                delta_reg_scores, datacube = self.calculate_ISDs(subset_accessibility, subset_factor_binding, subset_rp_map)
 
                 self.log.append('Calculating p-values ...')
                 
                 p_vals = self.get_delta_RP_p_value(delta_reg_scores, label_vector)
         
+            
+            self.log.append('Introspecting accessibility around factors ...')
+            self.factor_acc_z_scores = self.introsect_accessibility(self.rp_matrix, self.factor_gene_mask, dataset_mask,
+                self.chromatin_model.model.coef_)
+
             self.log.append('Done!')
 
         if debug:
